import os, sys, re, glob
from optparse import OptionParser
from datetime import datetime, time, timedelta

VERSION = (1,0,0,0)
DEFAULT_LOG_DIR = r"C:\Program Files\City of Heroes"

A_HERO = "HERO"
A_VILL = "VILLAIN"
A_PRAE = "PRAETOREAN"

DATE_TIME_REGEX = "\d\d-\d\d-\d\d\d\d \d\d:\d\d:\d\d"

loginHero     = re.compile(r"^(%s) Welcome to City of Heroes, (.*)!$" % DATE_TIME_REGEX)
loginVillain  = re.compile(r"^(%s) Now entering the Rogue Isles, (.*)!$" % DATE_TIME_REGEX)
loginEither   = re.compile(r"^(%s) (?:(?:Welcome to City of Heroes,)|(?:Now entering the Rogue Isles,)) (.*)!$"  % DATE_TIME_REGEX)
loginPraetor  = loginHero

infRegexes = ("^(%s) You gain ([0-9,]+) %%s" % DATE_TIME_REGEX,
              "^(%s) You work off [0-9,]+ debt and gain ([0-9,]+) %%s" % DATE_TIME_REGEX,
              "^(%s) You gain [0-9,]+ experience and ([0-9,]+) %%s" % DATE_TIME_REGEX,
              "^(%s) You gain [0-9,]+ experience, work off [0-9,]+ debt, and gain ([0-9,]+) %%s" % DATE_TIME_REGEX,)

commonPattern = re.compile(r"(%s) You received Invention: ([^)]+) \(Recipe\)" % DATE_TIME_REGEX)

dateFormats = ("%m/%d/%Y %H:%M:S",
               "%m/%d/%Y %H:%M",
               "%m/%d/%Y %H",
               "%m/%d/%Y",
               "%m/%d/%y %H:%M:S",
               "%m/%d/%y %H:%M",
               "%m/%d/%y %H",
               "%m/%d/%y")

commonPrices = {"Accuracy"            : 113650,
                "Confuse"             : 65025,
                "Damage"              : 116100,
                "Defense Buff"        : 81275,
                "Defense DeBuff"      : 84750,
                "Endurance Mod."      : 92875,
                "Endurance Red."      : 107975,
                "Fear"                : 53400,
                "Fly"                 : 73150,
                "Healing"             : 116100,
                "Hold Duration"       : 76625,
                "Immobilize Duration" : 84750,
                "Intangibility"       : 40650,
                "Interrupt Duration"  : 69650,
                "Jump"                : 46450,
                "Knockback Distance"  : 40650,
                "Range"               : 92875,
                "Recharge Red."       : 107975,
                "Resist Damage"       : 76625,
                "Run"                 : 76625,
                "Sleep Duration"      : 84750,
                "Snare Duration"      : 84750,
                "Stun Duration"       : 88225,
                "Taunt Duration"      : 58050,
                "To Hit Buff"         : 84750,
                "ToHit DeBuff"        : 61525,}

usage = "Usage: %prog [options] \"MM/DD/YYYY [HH[:MM[:SS]]]\" \"MM/DD/YY[YY] [HH[:MM[:SS]]]\""
parser = OptionParser(usage)
parser.add_option("-i", "--interval", dest="interval", default="15",
                  help="Time interval for parsing, in minutes.", choices=("5","10","15","20","30","60"))
parser.add_option("--villain", dest="alignment", default=None, action="store_const", const=A_VILL,
                  help="Set aligment of character to parse for to villain.")
parser.add_option("--hero", dest="alignment", action="store_const", const=A_HERO,
                  help="Set aligment of character to parse for to hero.")
parser.add_option("--praetorean", dest="alignment", action="store_const", const=A_PRAE,
                  help="Set aligment of character to parse for to praetorean.")
parser.add_option("-n", "--name", dest="name", type="string", default=None,
                  help="Name of the character to parse data for.")
parser.add_option("-o", "--outfile", dest="outfile", type="string", default="Rates.csv",
                  help="Filename of the output file to create.")
parser.add_option("-g", "--game-dir", dest="gamedir", type="string", default=DEFAULT_LOG_DIR,
                  help="Location of the City of Heroes chat logs.")
parser.add_option("-v", "--version", dest="version", action="store_true", 
                  help="Print the version number for the application.")
opts, args = parser.parse_args()

if (opts.version):
  print("\nVersion: %s" % ".".join((str(subVer) for subVer in VERSION)))
  sys.exit(0)

interval = int(opts.interval)
if (interval not in (5,10,15,20,30,60)):
  parser.print_help()
  sys.exit(1)

if (len(args) != 2):
  parser.print_help()
  sys.exit(1)

startTime = None
for format in dateFormats:
  try:
    startTime = datetime.strptime(args[0],format)
  except ValueError:
    continue
if (not startTime):
  print("You must provide a start time in the format MM/DD/YY[YY] [HH[:MM[:SS]]]")
  sys.exit(1)

endTime = None
for format in dateFormats:
  try:
    endTime = datetime.strptime(args[1],format)
  except ValueError:
    continue
if (not endTime):
  print("You must provide an end time in the format MM/DD/YY[YY] [HH[:MM[:SS]]]")
  sys.exit(1)

# Calculate all the log file names in the date range given.
oneDay = timedelta(days=1)
logDate = startTime.date()
endLogDate = endTime.date()

logFiles = []
while(logDate <= endLogDate):
  logFiles.append("chatlog %s.txt" % logDate.strftime("%m-%d-%Y"))
  logDate += oneDay
print("The following log files will be processed if they exist" \
      ".\n%s\n" % "\n".join("  %s" % logFile for logFile in logFiles))

# Calculate the datetime objects representing one hour divided into the requested intervals.
offsets = tuple([timedelta(0,0,0,0,step*interval) for step in range(0,int(60/interval))])

# Set up the regular expressions based on what formmat was requested.
if (opts.alignment is A_HERO):
  infName = "influence"
  loginPattern = loginHero
elif (opts.alignment is A_VILL):
  infName = "infamy"
  loginPattern = loginVillain
elif (opts.alignment is A_PRAE):
  infName = "information"
  loginPattern = loginPraetor
else:
  infName = "inf"
  loginPattern = loginEither

infPatterns = tuple(re.compile(regex % infName) for regex in infRegexes)

# Create a large list with all the log data, one line per list entry  
lines = []
for logFile in logFiles:
  logFileName = "%s\logs\game\%s" % (opts.gamedir,logFile)
  print("Reading %s..." % logFileName)
  if (not os.path.exists(logFileName)):
    print("WARNING: File '%s' does not exist." % logFileName)
    continue
  try:
    f = open(logFileName)
    lines.extend(f.readlines())
    f.close()
  except Exception as e:
    print("ERROR: Could not read log file '%s'.")
    print(e)
print("A total of %i log lines were read.\n" % len(lines))

print("Scanning and parsing log entries...")
rewardData = {}
name = opts.name
correctCharacter = False

if (len(offsets) > 1): offset = offsets[1]
else: offset = timedelta(hours=1)

# Iterate over all the lines, and match them against our patterns.
for count,line in enumerate(lines):
  if (not (count % 1000)):
    sys.stdout.write("Processed %i lines...\r" % count)
  recipe = False
  # If we set a character name on the command line...
  if (name):
    # ...check for login messages.
    match = loginPattern.search(line)
    # If this was a login message, determine if it's the character we want.
    if (match):
      correctCharacter = (match.group(2) == name)
    # If we're checking names and the last name seen wasn't our character
    # just skip lines until we see the name we want.
    if (not correctCharacter): continue
    
  # Check each inf pattern match to see if this is an inf reward line.
  for infPattern in infPatterns:
    match = infPattern.search(line)
    if (match): break
  if (match):
    timestamp,amount = match.groups()
  else:
    # If it wasn't an inf line, see if it was a recipe line.
    match = commonPattern.search(line)
    if (match):
      # If it was, extract the recipe name and look up its vendor value.
      timestamp,recipe =  match.groups()
      amount = commonPrices[recipe]
      recipe = True

  # If we had a reward match, calculate what interval it should be recorded in.
  if (match):
    month = int(timestamp[0:2])
    day = int(timestamp[3:5])
    year = int(timestamp[6:10])
    hour = int(timestamp[11:13])
    mins = int(timestamp[14:16])
    secs = int(timestamp[17:18])
    
    # Check this log entry against our start and end times.
    # If it's outside the range, then skip it.
    dt = datetime(year,month,day,hour,mins,secs)
    if (not (startTime <= dt <= endTime)): continue

    # If this is a string right from a reward line, turn it into an integer.
    if (isinstance(amount,str)):
      amount = int(amount.replace(',',''))
    
    # This is the nearest even interval for that reward time/datestamp
    nearest = datetime(year,month,day,hour) + offsets[int(mins/interval)]

    # If this was a recipe, record it in that data set.
    # Otherwise record it in the raw inf data set.
    defeats,recipes = rewardData.setdefault(nearest,([],[]))
    if (recipe):
      recipes.append(int(amount))
    else:
      defeats.append(int(amount))

print("Processed %i lines.  " % len(lines))

if (not rewardData):
  print("WARNING: No data found.")
  sys.exit(0)

# Open the output file, or die.
print ("Creating output file '%s'..." % opts.outfile)
try:
  f = open(opts.outfile,'w')
except Exception as e:
  print("ERROR: Could not open file '%s'" % opts.outfile)
  print(e)
  sys.exit(1)

base = datetime(startTime.year,startTime.month,startTime.day,startTime.hour) + offsets[int(startTime.minute/interval)]
if (endTime.hour == endTime.minute == endTime.second == 0):
  limit = datetime(endTime.year,endTime.month,endTime.day) + timedelta(days=1)
else:
  limit = datetime(endTime.year,endTime.month,endTime.day,endTime.hour) + offsets[int(endTime.minute/interval)]

# Start at the first interval, then step one interval at a time until we get
# to the last interval we processed. This will create entries in the CSV for
# lines that have no corresponding log entries, allowing graphs to correctly
# represent time when we were inactive or not logged in.
interval = base
while (interval < limit):
  defeatRewards,recipeRewards = rewardData.get(interval,([],[]))
  num = len(defeatRewards)
  gains = sum(defeatRewards)
  recipes = sum(recipeRewards)
  print("%s,%i,%i" % (interval.strftime("%m/%d/%Y %H:%M:%S"),gains,recipes),file=f)
  interval += offset

# Close the output file  
f.close()
print ("Done!")